/**************************************************************************/
/* FindRef.c - actual look-up routines. FindRef() is the interface.       */
/**************************************************************************/

#include "FetchRefs.h"

/// Prototypes and variable declarations
static BOOL RefMatch(STRPTR ref);
static LONG PutRef(struct RefsEntry *entr);
static LONG FindRefInteractive(void);
static void BuildList(void);
static LONG FileLength(STRPTR filename);

static LONG NumOfRefs;
static STRPTR pattern;
static struct List RefsFound;

struct OneItem {

/* ln_name in the Node points to the string that is to be shown in the
 * listview. ln_type is 1 for file references, 2 for all else.
 *
 * refsentry is a pointer to the reference to fetch.
 *
 * ListViewName is used if build a new string for the listview in.
 */

    struct Node node;
    struct RefsEntry *refsentry;
    UBYTE ListViewName[];
};

struct SearchOptions SOpts;
///

/// Timing functions. Useful when doing benchmarks.
/*
struct DateStamp StartTime, StopTime;

void
StartTimer(void)
{
    DateStamp(&StartTime);
}

void
StopTimer(void)
{
    ULONG diff;

    DateStamp(&StopTime);
    diff = ((StopTime.ds_Days   - StartTime.ds_Days) * 24 * 60 * 60 * 50)
         + ((StopTime.ds_Minute - StartTime.ds_Minute)         * 60 * 50)
         + ( StopTime.ds_Tick   - StartTime.ds_Tick                     );
    Printf("Timer was on for %ld ticks = %ld.%02ld seconds\n", diff, diff / 50, (diff % 50) * 2);
}
*/
///

/// FindRef() - interface to fetch a reference to file
/* Actual look-up routine. Copies the reference to 'ref' into a file.
 * Arguments are passed through the global structure SOpts.
 */
LONG
FindRef(void)
{
    struct Node *f;
    LONG ret, len;

    /* Tokenize pattern string */
    len = strlen(SOpts.Reference) * 2 + 2;

    if (pattern = AllocVec(len, MEMF_CLEAR))
    {
        LONG ret;

        if (SOpts.Case)
            ret = ParsePattern(SOpts.Reference, pattern, len);
        else
            ret = ParsePatternNoCase(SOpts.Reference, pattern, len);

        if (ret == -1)
        {
            FreeVec(pattern);
            return(ERROR_TOO_MANY_LEVELS);
        }
    } else
        return(ERROR_NO_FREE_STORE);
    
    /* Build a list of references matching the requested one */
    BuildList();

    if (NumOfRefs > 1)
        ret = FindRefInteractive();
    else if (NumOfRefs == 1)
        ret = PutRef(((struct OneItem *)RefsFound.lh_Head)->refsentry);
    else
        ret = 1;

    /* Free the list */
    while (f = RemHead(&RefsFound))
        FreeVec(f);

    FreeVec(pattern);

    return(ret);
}
///

/// RefMatch(STRPTR ref) - Check if a reference name is matching the requested
static BOOL
RefMatch(STRPTR ref)
{
    if (SOpts.Case)
        return(MatchPattern(pattern, ref));
    else
        return(MatchPatternNoCase(pattern, ref));
}
///

/// PutRef(struct RefsEntry *entr) - writes a reference to file
/* Write the reference. Handles the writing both to a normal file and to the
 * clipboard (through iffparse.library).
 */
#define ID_FTXT MAKE_ID('F','T','X','T')
#define ID_CHRS MAKE_ID('C','H','R','S')
static LONG
PutRef(struct RefsEntry *entr)
{
    BPTR from;
    STRPTR c = NULL;
    LONG result = 1;

    /* Figure out the actual length if the reference is a complete file */
    if (entr->Length == -1)
        entr->Length = FileLength(entr->node.ln_Name);
    
    if (from = Open(entr->node.ln_Name, MODE_OLDFILE))
    {
        if (c = AllocVec(entr->Length, NULL))
        {
            Seek(from, entr->Offset, OFFSET_BEGINNING);
            Read(from, c, entr->Length);
        } else
            result = ERROR_NO_FREE_STORE;
        Close(from);
    } else
        result = IoErr();

    if (c)
    {
        if ((strlen(SOpts.DestFile) <= 4) || (strncmp(SOpts.DestFile, "CLIP", 4) != 0))
        {
            BPTR to;
            if (to = Open(SOpts.DestFile, MODE_NEWFILE))
            {
                if (Write(to, c, entr->Length) != entr->Length)
                    result = IoErr();
                else
                    result = 0;
                Close(to);
            } else
                result = IoErr();
        } else
        {
            struct Library *IFFParseBase;
            if (IFFParseBase = OpenLibrary("iffparse.library", 37))
            {
                struct IFFHandle *iff;

                if (iff = AllocIFF())
                {
                    if (iff->iff_Stream = OpenClipboard(atoi(&(SOpts.DestFile[4]))))
                    {
                        InitIFFasClip(iff);
                        if (OpenIFF(iff, IFFF_WRITE) == 0)
                        {
                            if (!PushChunk(iff, ID_FTXT, ID_FORM, IFFSIZE_UNKNOWN))
                            {
                                if (!PushChunk(iff, 0, ID_CHRS, entr->Length))
                                {
                                    if (WriteChunkBytes(iff, c, entr->Length) == entr->Length)
                                    {
                                        result = 0;
                                        PopChunk(iff);
                                        PopChunk(iff);
                                    }
                                }
                            }
                            CloseIFF(iff);
                        }
                        CloseClipboard((struct ClipboardHandle *)iff->iff_Stream);
                    }
                    FreeIFF(iff);
                }
                CloseLibrary(IFFParseBase);
            }
        }
        FreeVec(c);
    }

    if (result == 0)
        GotoLine = entr->Goto;

    return(result);
}
///

/// BuildList() - builds a list of all the matching references
/* This is a call back function for the sort routine. Puts files in top. */
BOOL
SortListFunc(struct OneItem *a, struct OneItem *b, ULONG data)
{
    if (a->node.ln_Type == b->node.ln_Type)
        return( stricmp(a->node.ln_Name, b->node.ln_Name) > 0 );

    return(a->node.ln_Type == 2);
}

void
BuildList(void)
{
    struct RefsEntry *currref;
    struct FileEntry *currfile;

    /* First of all: Initialize the list */
    NewList(&RefsFound);
    NumOfRefs = 0;

    /* Pick up all references */
    for (currfile = FileList.lh_Head; currfile->node.ln_Succ; currfile = currfile->node.ln_Succ)
    {
        /* Check for a match on the file name */
        if (SOpts.FileRef)
        {
            UBYTE name[108], *cutptr;

            /* Isolate the base name (fx 'INCLUDE:libraries/dos.h' -> 'dos') */
            strcpy(name, FilePart(currfile->Name));
            cutptr = name;
            while (*cutptr)
            {
                if ((*cutptr >= 'A' && *cutptr <= 'Z') ||
                    (*cutptr >= 'a' && *cutptr <= 'z') ||
                    (*cutptr >= '0' && *cutptr <= '9') ||
                    (*cutptr == '_'))
                    cutptr++;
                else
                    *cutptr = 0;
            }

            /* Generate an item to add to the RefsFound list */
            if (RefMatch(name))
            {
                struct {
                    struct OneItem item;
                    struct RefsEntry refs;
                } *Thing;

                if (Thing = AllocVec(sizeof(struct OneItem) + sizeof(struct RefsEntry), MEMF_CLEAR))
                {
                    Thing->item.refsentry = &Thing->refs;
                    Thing->item.node.ln_Type = 1;
                    Thing->item.node.ln_Name = currfile->Name;

                    Thing->refs.node.ln_Name = currfile->Name;
                    Thing->refs.Length = -1; /* -1 means "the whole file" */

                    AddTail(&RefsFound, &Thing->item.node);
                    NumOfRefs++;
                }
            }
        }

        /* Check all the RefsEntry's in this FileEntry for a match */
        for (currref = currfile->RefsList.lh_Head; currref->node.ln_Succ; currref = currref->node.ln_Succ)
        {
            if (RefMatch(currref->Name))
            {
                struct OneItem *r;

                if (r = AllocVec(sizeof(struct OneItem) + strlen(currfile->Name) + strlen(currref->Name) + 5, MEMF_CLEAR))
                {
                    r->refsentry = currref;
                    r->node.ln_Type = 1;
                    r->node.ln_Name = r->ListViewName;
                    SPrintf(r->ListViewName, "%s in %s", currref->Name, currfile->Name);

                    AddTail(&RefsFound, &r->node);
                    NumOfRefs++;
                }
            }
        }
    }
    SortExecList(&RefsFound, SortListFunc, 0);
}
///

/// FileLength(STRPTR filename) - returns file length or -1 for failure
LONG
FileLength(STRPTR name)
{
    LONG length = -1;
    BPTR lock;
    struct FileInfoBlock *file;

    if (file = AllocVec(sizeof(struct FileInfoBlock), MEMF_CLEAR))
    {
        if (lock = Lock(name, ACCESS_READ))
        {
            Examine(lock, file);
            UnLock(lock);
            length = file->fib_Size;
        }
        FreeVec(file);
    }
    return(length);
}
///

/// FindRefInteractive() - open window and ask
/*
 *  Source machine generated by GadToolsBox V2.0b
 *  which is (c) Copyright 1991-1993 Jaba Development
 *
 *  GUI Designed by : Anders Melchiorsen
 *
 *  The following is somewhat based on the output from GadToolsBox but it has
 *  been changed considerably to fit better here.
 */

struct NewGadget TheListGadget;

struct TagItem GadgetTags[] = {
    { (GTLV_Labels), &RefsFound },
    { (TAG_DONE)                }
};

/* Build a List of all files containing the reference, prompt for a selection
 * and get it. This is a quite long function, but I have a few reasons if you
 * start complaining... However, I do admit: It does look a little clumsy :-)
 */
LONG
FindRefInteractive()
{
    APTR VisualInfo = NULL;
    struct Screen *Scr = NULL;
    struct Window *Wnd = NULL;
    struct Gadget *GadgetList = NULL;
    LONG count = 0, click = -1, result = 1;
    BOOL poppedscr = FALSE;
    UWORD WndLeft, WndWidth, WndTop, WndHeight;
    struct IntuiMessage *m;

    /* Find a screen pointer to open on */
    if (SOpts.PubScreen)
        Scr = LockPubScreen((STRPTR)SOpts.PubScreen);
    if (!Scr)
    {
        UBYTE wantname[MAXPUBSCREENNAME+1];
        struct Screen *wantscr;
        struct List *publist;
        struct PubScreenNode *pubnode;

        publist = LockPubScreenList();
        wantscr = IntuitionBase->ActiveScreen;
        pubnode = (struct PubScreenNode *)publist->lh_Head;
        while (pubnode->psn_Node.ln_Succ)
        {
            if (pubnode->psn_Screen == wantscr)
                strcpy(wantname, pubnode->psn_Node.ln_Name);
            pubnode = pubnode->psn_Node.ln_Succ;
        }
        UnlockPubScreenList();
        Scr = LockPubScreen(wantname);
    }
    if (!Scr)
        Scr = LockPubScreen(NULL);

    /* Create gadget, open window */
    if ((!Scr) || !(VisualInfo = GetVisualInfo(Scr, TAG_DONE)) || !(CreateContext(&GadgetList)))
        goto getout;

    TheListGadget.ng_VisualInfo = VisualInfo;
    TheListGadget.ng_TextAttr = Scr->Font;

    /* Now pretend to be font sensitive :-) */

    WndTop = 3 * (Scr->WBorTop + Scr->RastPort.TxHeight + 1);
    WndHeight = (NumOfRefs + 2) * Scr->RastPort.TxHeight + 4;
    if (WndHeight > Scr->Height / 2)
        WndHeight = Scr->Height / 2;
    WndLeft = 5 * Scr->RastPort.TxWidth - Scr->WBorLeft;
    WndWidth = Scr->Width - (2 * WndLeft);

    TheListGadget.ng_TopEdge  = Scr->RastPort.TxHeight;
    TheListGadget.ng_Height   = WndHeight - (2 * Scr->RastPort.TxHeight);
    TheListGadget.ng_LeftEdge = Scr->RastPort.TxWidth;
    TheListGadget.ng_Width    = WndWidth - (2 * TheListGadget.ng_LeftEdge);

    if (!(CreateGadgetA(LISTVIEW_KIND, GadgetList, &TheListGadget, GadgetTags)))
        goto getout;

    /* Open window */
    if (Wnd = OpenWindowTags( NULL,
            WA_Left,        WndLeft,
            WA_Top,         WndTop,
            WA_InnerWidth,  WndWidth,
            WA_InnerHeight, WndHeight,
            WA_IDCMP,       IDCMP_VANILLAKEY|LISTVIEWIDCMP|IDCMP_CLOSEWINDOW|IDCMP_REFRESHWINDOW,
            WA_Flags,       WFLG_ACTIVATE|WFLG_DRAGBAR|WFLG_DEPTHGADGET|WFLG_CLOSEGADGET|WFLG_SMART_REFRESH|WFLG_RMBTRAP|WFLG_GIMMEZEROZERO,
            WA_Gadgets,     GadgetList,
            WA_Title,       "Please select reference",
            WA_ScreenTitle, "FetchRefs " VERSION " by Anders Melchiorsen",
            WA_PubScreen,   Scr,
            TAG_DONE ))
    {
        GT_RefreshWindow( Wnd, NULL );

        if (Scr != IntuitionBase->FirstScreen)
        {
            poppedscr = TRUE;
            ScreenToFront(Scr);
        }
    } else
        goto getout;

    /* Handle IDCMP */
    while (click == -1)
    {
        WaitPort(Wnd->UserPort);
        while( m = GT_GetIMsg( Wnd->UserPort ))
        {
            switch ( m->Class )
            {
                case IDCMP_REFRESHWINDOW:
                    GT_BeginRefresh( Wnd );
                    GT_EndRefresh( Wnd, TRUE );
                break;

                case IDCMP_VANILLAKEY:
                    if ( m->Code != 27)
                        break;
                    /* Fall through to CLOSEWINDOW only if ESC was pressed */

                /* Check for a closed window (abort) */
                case IDCMP_CLOSEWINDOW:
                    click = -2;
                    result = 2;
                break;

                /* Check for double click */
                case IDCMP_GADGETUP:
                {
                    static ULONG lastsec, lastmic, lastclick = -1;

                    if (lastclick == m->Code && DoubleClick(lastsec, lastmic, m->Seconds, m->Micros))
                        click = m->Code;
                    lastclick = m->Code;
                    lastsec = m->Seconds;
                    lastmic = m->Micros;
                }
                break;
            }
            GT_ReplyIMsg( m );
        }
    }

getout:
    if (poppedscr)
        ScreenToBack(Scr);

    if (Wnd)
        CloseWindow(Wnd);

    if (Scr)
        UnlockPubScreen(NULL, Scr);

    if (GadgetList)
        FreeGadgets(GadgetList);

    if (VisualInfo)
        FreeVisualInfo(VisualInfo);

    /* Find the selected Node and put the reference (if a double click
     * occoured at all).
     */
    if (click > -1)
    {
        struct Node *f;

        f = RefsFound.lh_Head;
        while (click--)
            f = f->ln_Succ;
        result = PutRef(((struct OneItem *)f)->refsentry);
    }

    /* (The RefsFound list is freed by the calling FindRef) */

    return(result);
}
///

